"""
======
meltPT
======

A Python package for whole-rock major-element themormobarometric analyses of basaltic (mafic) rocks.

"""


from .parse import *
from .backtrack_compositions import *
from .thermobarometry import *
from .fit_melting_paths import *
from .suite import *